from lightgbm import LGBMClassifier, Booster
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
from sklearn.metrics import accuracy_score, f1_score, precision_recall_fscore_support
import json
import numpy as np

def construct_data(perplexity_file):
    perplexity = []
    label = []
    with open(perplexity_file, "r", encoding="utf-8") as f:
        for line in f:
            try:
                data = json.loads(line)
                for i, j in data.items():
                    perplexity.append(j)
                    if i == "human_write":
                        label.append(0)
                    elif i == "gpt2":
                        label.append(1)
                    elif i == "opt":
                        label.append(2)
                    elif i == "unilm":
                        label.append(3)
            except Exception as e:
                print(e, line)

    return perplexity, label

def LightGBM_Classification(X, y, n):
    # Load train dataset
    X = np.array(X)
    y = np.array(y)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=12343)

    model = LGBMClassifier(
        max_depth=3,
        learning_rate=0.1,
        n_estimators=200,
        objective='multiclass',
        num_class=n,
        booster='gbtree',
        min_child_weight=2,
        subsample=0.8,
        colsample_bytree=0.8,
        reg_alpha=0,
        reg_lambda=1,
        seed=0)

    model.fit(X_train, y_train, eval_set=[(X_train, y_train), (X_test, y_test)], verbose=100, early_stopping_rounds=50)

    # Save model
    model.booster_.save_model("LightGBM_model.txt")

    return X_test,  y_test

if __name__ == "__main__":
    GPT_perplexity, label_1 = construct_data("new_gpt_ppl.json")
    OPT_perplexity, label_2 = construct_data("new_opt_ppl.json")
    UniLM_perplexity, label_3 = construct_data("new_unilm_ppl.json")
    perplexity = [[i, j, k] for i, j, k in zip(GPT_perplexity, OPT_perplexity, UniLM_perplexity)]

    # Ground truth label
    ground_label = []
    for i, j, k in zip(label_1, label_2, label_3):
        if i == j and i == k and j == k:
            ground_label.append(i)

    # Train classification model and return test data
    X_test, y_test = LightGBM_Classification(perplexity, ground_label, 4)

    # Load trained model
    model = Booster(model_file='LightGBM_model.txt')
    # Make predictions on the test setMake predictions on the test set
    y_pred = model.predict(X_test)
    y_pred = [np.argmax(i) for i in y_pred]
    # Calculate accuracy, recall, f1 value
    p_class, r_class, f_class, support_micro = precision_recall_fscore_support(y_test, y_pred, labels=[i for i in range(4)])
    print(f_class)




import os
import json
import argparse
from functools import partial
from multiprocessing import Pool  # 多线程并行
import tqdm
import torch
import pickle
import torch.nn.functional as F
from nltk import ngrams
from collections import Counter
import math
import random

# mutil-GPU
from torch.nn.parallel import DistributedDataParallel as DDP
import torch.distributed as dist

from transformers import HfArgumentParser
from transformers import AutoTokenizer, LlamaTokenizer, AutoModelForSeq2SeqLM, AutoModelForCausalLM, LlamaForCausalLM

from unilm import UniLMTokenizer, UniLMForConditionalGeneration

from arguments import parse_args

# Load and return model and tokenzier
def load_model(args):
    # Load corresponding model and tokenizer according to  model name ot path
    if args.is_seq2seq_model:
        tokenizer = AutoTokenizer.from_pretrained(args.model_name_or_path)
        model = AutoModelForSeq2SeqLM.from_pretrained(args.model_name_or_path)
    elif args.is_decoder_only_model:
        tokenizer = AutoTokenizer.from_pretrained(args.model_name_or_path)
        model = AutoModelForCausalLM.from_pretrained(args.model_name_or_path)
    elif args.is_llama:
        tokenizer = LlamaTokenizer.from_pretrained(args.model_name_or_path)
        model = LlamaForCausalLM.from_pretrained(args.model_name_or_path, torch_dtype=torch.float16)
    elif args.is_unilm:
        tokenizer = UniLMTokenizer.from_pretrained(args.model_name_or_path)
        model = UniLMForConditionalGeneration.from_pretrained(args.model_name_or_path)
    else:
        raise ValueError(f"Unknown model type: {args.model_name_or_path}")

    if args.use_gpu:
        device = "cuda" if torch.cuda.is_available() else "cpu"
        model = model.to(device)
    else:
        device = "cpu"

    return tokenizer, model, device

# Generate text according to prompt and model
def generate(input, args, tokenizer, model, device):
    gen_kwargs = dict(max_new_tokens=args.max_new_tokens)
    if args.use_sampling:
        gen_kwargs.update(dict(
            do_sample=True,
            top_k=0,
            temperature=args.sampling_temp
        ))
    else:
        gen_kwargs.update(dict(
            num_beams=args.n_beams
        ))
    model = model.to(device)
    # generation function
    generate = partial(model.generate, **gen_kwargs)

    # restrict the length of input
    if args.prompt_max_length:
        pass
    elif hasattr(model.config, "max_position_embedding"):
        args.prompt_max_length = model.config.max_position_embeddings - args.max_new_tokens
    else:
        args.prompt_max_length = 2048 - args.max_new_tokens

    tokd_input = tokenizer(input, return_tensors="pt", add_special_tokens=True, truncation=True,
                           max_length=args.prompt_max_length).to(device)

    torch.manual_seed(args.generation_seed)
    output = generate(**tokd_input)

    # need to isolate the newly generated tokens
    output = output[:, tokd_input["input_ids"].shape[-1]:]

    decoded_output_text = tokenizer.batch_decode(output, skip_special_tokens=True)[0]

    return decoded_output_text

# Used to count the top k n-grams  in a given text set
def count_n_grams(text_set, n, top_k, tokenizer):
    n_grams = []
    for text in text_set:
        text_n_grams = ngrams(tokenizer.convert_tokens_to_ids(tokenizer.tokenize(text)), n)
        n_grams.extend(text_n_grams)

    # Count and sort(according to value in dict, return key) n-grams，
    n_grams = sorted(Counter(n_grams), reverse=True)

    if len(n_grams) < top_k:
        return n_grams
    else:
        return n_grams[:top_k]

def next_token_probs(n_grams, top_k, model, tokenizer, device):
    grams_probs = {}
    for i in tqdm.tqdm(range(len(n_grams))):
        prompt = tokenizer.decode(n_grams[i])
        tokd_input = tokenizer.encode(prompt, return_tensors="pt").to(device)

        model.to(device)
        with torch.no_grad():
            outputs = model(tokd_input)
            predictions = outputs[0]

        logits = predictions[0, -1, :]

        # probability
        probs = F.softmax(logits, dim=0)

        values, indices = torch.topk(probs, top_k, -1)
        indices = indices.tolist()
        values = values.tolist()
        grams_probs[str(n_grams[i])] = dict(zip(indices, values))

    return grams_probs

# construct test text set with original text set and generated test set
def construct_test_text_set(original_text_set, generated_test_set, args):
    test_set = []
    for i in generated_test_set:
        if i != "":
            test_set.append({"text": i, "label": args.model_name})
    test_set.extend(original_text_set)
    random.shuffle(test_set)
    return test_set

# Calculate proxy perplexity according to n-grams probability
def proxy_perplexity(text_tokenize_ids, args, probability_2_grams, probability_3_grams, probability_4_grams, probability_5_grams):
    test_ppl = []
    # Calculate proxy perplexity
    for j in tqdm.tqdm(range(len(text_tokenize_ids))):
        for label, tokenize_ids in text_tokenize_ids[j].items():
            ppl_result = {}
            # ppl = math.log2(probability_2_grams[tokenize_ids[0]][str(tokenize_ids[1])])
            ppl = 0
            number_5_grams = 0
            number_4_grams = 0
            number_3_grams = 0
            number_2_grams = 0
            for i in range(3, len(tokenize_ids)-1):

                # 5-grams proxy perplexity
                if str([tokenize_ids[i - j] for j in range(3, -1, -1)]) in probability_5_grams.keys():
                    if str(tokenize_ids[i + 1]) in probability_5_grams[str([tokenize_ids[i - j] for j in range(3, -1, -1)])].keys():
                        ppl = ppl + math.log2(probability_5_grams[str([tokenize_ids[i - j] for j in range(3, -1, -1)])][str(tokenize_ids[i + 1])])
                    else:
                        top_k = len(probability_5_grams[str([tokenize_ids[i - j] for j in range(3, -1, -1)])].keys())
                        sum_probs = sum(probability_5_grams[str([tokenize_ids[i - j] for j in range(3, -1, -1)])].values())
                        if (1 - sum_probs) > 0:
                            ppl = ppl + math.log2((1 - sum_probs) / (args.vocab_length - top_k))
                    number_5_grams = number_5_grams + 1

                # 4-grams proxy perplexity
                elif str([tokenize_ids[i - j] for j in range(2, -1, -1)]) in probability_4_grams.keys():
                    if str(tokenize_ids[i + 1]) in probability_4_grams[str([tokenize_ids[i - j] for j in range(2, -1, -1)])].keys():
                        ppl = ppl + math.log2(probability_4_grams[str([tokenize_ids[i - j] for j in range(2, -1, -1)])][str(tokenize_ids[i + 1])])
                    else:
                        top_k = len(probability_4_grams[str([tokenize_ids[i - j] for j in range(2, -1, -1)])].keys())
                        sum_probs = sum(probability_4_grams[str([tokenize_ids[i - j] for j in range(2, -1, -1)])].values())
                        if (1 - sum_probs) > 0:
                            ppl = ppl + math.log2((1 - sum_probs) / (args.vocab_length - top_k))
                    number_4_grams = number_4_grams + 1

                # 3-grams proxy perplexity
                elif str([tokenize_ids[i - 1], tokenize_ids[i]]) in probability_3_grams.keys():
                    if str(tokenize_ids[i + 1]) in probability_3_grams[str([tokenize_ids[i - 1], tokenize_ids[i]])].keys():
                        ppl = ppl + math.log2(probability_3_grams[str([tokenize_ids[i - 1], tokenize_ids[i]])][str(tokenize_ids[i + 1])])
                    else:
                        top_k = len(probability_3_grams[str([tokenize_ids[i - 1], tokenize_ids[i]])].keys())
                        sum_probs = sum(probability_3_grams[str([tokenize_ids[i - 1], tokenize_ids[i]])].values())
                        if (1 - sum_probs) > 0:
                            ppl = ppl + math.log2((1 - sum_probs) / (args.vocab_length - top_k))
                    number_3_grams = number_3_grams + 1

                # 2-grams proxy perplexity
                elif str([tokenize_ids[i]]) in probability_2_grams.keys():
                    if str(tokenize_ids[i+1]) in probability_2_grams[str([tokenize_ids[i]])].keys():
                        ppl = ppl + math.log2(probability_2_grams[str([tokenize_ids[i]])][str(tokenize_ids[i+1])])
                    else:
                        top_k = len(probability_2_grams[str([tokenize_ids[i]])].keys())
                        sum_probs = sum(probability_2_grams[str([tokenize_ids[i]])].values())
                        ppl = ppl + math.log2((1-sum_probs) / (args.vocab_length - top_k))
                    number_2_grams = number_2_grams + 1
            ppl = round(ppl / (number_2_grams + number_3_grams + number_4_grams + number_5_grams + 1), 2)
            ppl_result[label] = -ppl
            test_ppl.append(ppl_result)
    return test_ppl

def main(args):
    args.is_seq2seq_model = any([(model_type in args.model_name_or_path) for model_type in ["t5", "T0"]])
    args.is_decoder_only_model = any([(model_type in args.model_name_or_path) for model_type in ["gpt", "opt", "bloom"]])
    args.is_gpt = any([(model_type in args.model_name_or_path) for model_type in ["gpt"]])
    args.is_opt = any([(model_type in args.model_name_or_path) for model_type in ["opt"]])
    args.is_unilm = any([(model_type in args.model_name_or_path) for model_type in ["unilm"]])
    args.is_llama = any([(model_type in args.model_name_or_path) for model_type in ["llama"]])

    if args.is_gpt:
        args.model_name = "gpt"
        args.vocab_length = 50257
    elif args.is_opt:
        args.model_name = "opt"
        args.vocab_length = 50265
    elif args.is_unilm:
        args.model_name = "unilm"
        args.vocab_length = 28996
    elif args.is_llama:
        args.model_name = "llama"
        args.vocab_length = 32000

    if args.prompt_file is not None:
        extension = args.prompt_file.split(".")[-1]
        assert extension in ["csv", "json", "tsv"], "`prompt_file` should be a csv or a json file."

    # load prompt set
    prompt = []
    with open(args.prompt_file, 'r', encoding='utf-8') as f:
        for line in f:
            try:
                text = json.loads(line)
                prompt.append(text)

            except Exception as e:
                print(e, line)

    # load model and tokenizer
    torch.cuda.set_device(args.local_rank)
    dist.init_process_group(backend='nccl')
    tokenizer, model, device = load_model(args)
    model = DDP(model)

    generate_fn = partial(generate, args=args, tokenizer=tokenizer, model=model, device=device)
    generation_text = []

    # check if there is a saved state
    if args.save_state:
        try:
            with open(args.save_state, 'rb') as f:
                state = pickle.load(f)
                generation_text = state['generation_text']
                start_batch = state['start_batch']

        except Exception as e:
            print(e)

    else:
        start_batch = 0

    for batch in tqdm.tqdm(range(start_batch, args.n_generate_text // (args.batch_size * dist.get_world_size()))):
        input = prompt[batch * args.batch_size * dist.get_world_size():(batch + 1) * args.batch_size * dist.get_world_size()]
        for idx in range(len(input)):
            generation_text.append(generate_fn(input[idx]))

        # save state after each batch
        if args.save_state:
            state = {
                "generateion_text": generation_text,
                "state_batch": batch - 1
            }
            with open(args.save_state, "wb") as f:
                pickle.dump(state, f)

    out_dir = os.path.join("result", args.model_name)
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    generation_text_file = os.path.join(out_dir, args.generated_text_file)
    with open(generation_text_file, "w", encoding='utf-8') as f:
        for text in generation_text:
            f.write(json.dumps(text, ensure_ascii=False))
            f.write('\n')

    # save final state
    if args.save_state:
        state = {
            'generation_text': generation_text,
            'start_batch': args.n_generate_text // args.batch_size,
        }
        with open(args.save_state, 'wb') as f:
            pickle.dump(state, f)

    statistics_text_set = generation_text[:int(args.n_generate_text * 0.8)]
    generation_test_text_set = generation_text[int(args.n_generate_text * 0.8):args.n_generate_text]

    # count the number of n-grams
    one_grams = count_n_grams(statistics_text_set, 1, args.n_one_grams, tokenizer)
    two_grams = count_n_grams(statistics_text_set, 2, args.n_two_grams, tokenizer)
    three_grams = count_n_grams(statistics_text_set, 3, args.n_three_grams, tokenizer)
    four_grams = count_n_grams(statistics_text_set, 4, args.n_four_grams, tokenizer)

    # samples the next token probability of n-grams
    two_grams_probs = next_token_probs(one_grams, args.two_grams_top_k, model, tokenizer, device)
    three_grams_probs = next_token_probs(two_grams, args.three_grams_top_k, model, tokenizer, device)
    four_grams_probs = next_token_probs(three_grams, args.four_grams_top_k, model, tokenizer, device)
    five_grams_probs = next_token_probs(four_grams, args.five_grams_top_k, model, tokenizer, device)

    two_grams_probs_file = os.path.join(out_dir, args.two_grams_probs_file)
    with open(two_grams_probs_file, 'w', encoding='utf-8') as f:
        f.write(json.dumps(two_grams_probs, ensure_ascii=False))

    three_grams_probs_file = os.path.join(out_dir, args.three_grams_probs_file)
    with open(three_grams_probs_file, 'w', encoding='utf-8') as f:
        f.write(json.dumps(three_grams_probs, ensure_ascii=False))

    four_grams_probs_file = os.path.join(out_dir, args.four_grams_probs_file)
    with open(four_grams_probs_file, 'w', encoding='utf-8') as f:
        f.write(json.dumps(four_grams_probs, ensure_ascii=False))

    five_grams_probs_file = os.path.join(out_dir, args.five_grams_probs_file)
    with open(five_grams_probs_file, 'w', encoding='utf-8') as f:
        f.write(json.dumps(five_grams_probs, ensure_ascii=False))

    original_text_set = []
    with open(args.original_test_file, 'r', encoding="utf-8") as f:
        for line in f:
            try:
                text = json.loads(line)
                original_text_set.append(text)
            except Exception as e:
                print(e, line)

    test_set = construct_test_text_set(original_text_set, generation_test_text_set, args)

    new_test_file = os.path.join(out_dir, args.new_test_file)
    with open(new_test_file, "w", encoding='utf-8') as f:
        for text in test_set:
            f.write(json.dumps(text, ensure_ascii=False))
            f.write('\n')

    test_text_tokenize_ids = []
    for i in tqdm.tqdm(range(len(test_set))):
        tokenize_ids = {}
        if args.is_gpt:
            text_tokenize = tokenizer.tokenize(test_set[i]["text"], truncation=True)
        else:
            text_tokenize = tokenizer.tokenize(test_set[i]["text"])
        tokenize_ids[text[i]["label"]] = tokenizer.convert_tokens_to_ids(text_tokenize)
        # tokens to token_ids
        test_text_tokenize_ids.append(tokenize_ids)

    test_ppl = proxy_perplexity(test_text_tokenize_ids, args, two_grams_probs, three_grams_probs, four_grams_probs, five_grams_probs)

    test_ppl_file = os.path.join(out_dir, args.test_ppl_file)
    with open(test_ppl_file, "w", encoding='utf-8') as f:
        for ppl in test_ppl:
            f.write(json.dumps(ppl, ensure_ascii=False))
            f.write('\n')



if __name__ == "__main__":
    args = parse_args()
    # print(args)
    main()




import argparse

def parse_args():
    """Command line argument specification"""
    parser = argparse.ArgumentParser(description="Arguments for Generation Text Detection")

    # Generation Arguments
    parser.add_argument(
        "--model_name_or_path",
        type=str,
        default="facebook/opt-1.3b",
        help="Main model, path to pretrained model or model identifier from huggingface.co/models.",
    )
    parser.add_argument(
        "--cache_dir",
        type=str,
        default=None,
        help="Path to directory to store the pretrained models downloaded from huggingface.co",
    )
    parser.add_argument(
        "--use_fast_tokenizer",
        type=bool,
        default=True,
        help="Whether to use one of the fast tokenizer (backed by the tokenizers library) or not.",
    )
    parser.add_argument(
        "--model_revision",
        type=str,
        default="main",
        help="The specific model version to use (can be a branch name, tag name or commit id).",
    )
    parser.add_argument(
        "--max_new_tokens",
        type=int,
        default=50,
        help="Maximum number of new tokens to generate.",
    )
    parser.add_argument(
        "--generation_seed",
        type=int,
        default=123,
        help="Seed for setting the torch global rng prior to generation.",
    )
    parser.add_argument(
        "--use_sampling",
        type=bool,
        default=True,
        help="Whether to generate using multinomial sampling.",
    )
    parser.add_argument(
        "--sampling_temp",
        type=float,
        default=0.7,
        help="Sampling temperature to use when generating using multinomial sampling.",
    )
    parser.add_argument(
        "--n_beams",
        type=int,
        default=1,
        help="Number of beams to use for beam search. 1 is normal greedy decoding",
    )
    parser.add_argument(
        "--use_gpu",
        type=bool,
        default=True,
        help="Whether to run inference and watermark hashing/seeding/permutation on gpu.",
    )
    parser.add_argument(
        "--prompt_max_length",
        type=int,
        default=None,
        help="Truncation length for prompt, overrides model config's max length field.",
    )
    parser.add_argument(
        "--prompt_file",
        type=str,
        default="text_generation/prompt.json",
        help="Path to prompt dataset for text generation.",
    )
    parser.add_argument(
        "--batch_size",
        type=int,
        default=16,
        help="The number of samples selected for text generation",
    )
    parser.add_argument(
        "--use_auth_token",
        type=bool,
        default=False,
        help="Will use the token generated when running `transformers-cli login` (necessary to use this script with private models).",
    )
    parser.add_argument(
        "--n_generate_text",
        type=int,
        default=96000,
        help="The numbers of generation text.",
    )

    # Samples Arguments
    parser.add_argument(
        "--n_one_grams",
        type=int,
        default=1000,
        help="The number of selected 1-grams in statistics text.",
    )
    parser.add_argument(
        "--n_two_grams",
        type=int,
        default=1000,
        help="The number of selected 2-grams in statistics text.",
    )
    parser.add_argument(
        "--n_three_grams",
        type=int,
        default=1000,
        help="The number of selected 3-grams in statistics text.",
    )
    parser.add_argument(
        "--n_four_grams",
        type=int,
        default=1000,
        help="The number of selected 4-grams in statistics text.",
    )
    parser.add_argument(
        "--two_grams_top_k",
        type=int,
        default=1000,
        help="The number of Top K samples of the next word in 3-grams.",
    )
    parser.add_argument(
        "--three_grams_top_k",
        type=int,
        default=1000,
        help="The number of Top K samples of the next word in 3-grams.",
    )
    parser.add_argument(
        "--four_grams_top_k",
        type=int,
        default=1000,
        help="The number of Top K samples of the next word in 4-grams.",
    )
    parser.add_argument(
        "--five_grams_top_k",
        type=int,
        default=1000,
        help="The number of Top K samples of the next word in 5-grams.",
    )
    parser.add_argument(
        "--two_grams_probs_file",
        type=str,
        default="two_grams_probs.json",
        help="The file of 2-grams probability.",
    )
    parser.add_argument(
        "--three_grams_probs_file",
        type=str,
        default="three_grams_probs.json",
        help="The file of 3-grams probability.",
    )
    parser.add_argument(
        "--four_grams_probs_file",
        type=str,
        default="four_grams_probs.json",
        help="The file of 4-grams probability.",
    )
    parser.add_argument(
        "--five_grams_probs_file",
        type=str,
        default="five_grams_probs.json",
        help="The file of 5-grams probability.",
    )

    # Proxy Perplexity Arguments
    parser.add_argument(
        "--original_test_file",
        type=str,
        default="test_set.json",
        help="The input testing data file.",
    )

    # Output File Arguments
    parser.add_argument(
        "--generated_text_file",
        type=str,
        default="generated_text.json",
        help="The output generated text file.",
    )
    parser.add_argument(
        "--new_test_file",
        type=str,
        default="new_test_set.json",
        help="The new test set file",
    )
    parser.add_argument(
        "--test_ppl_file",
        type=str,
        default="test_ppl.json",
        help="The result of Proxy Perplexity for test texts.",
    )
    args = parser.parse_args()
    return args





import numpy as np
from tqdm import tqdm
import math
import json
from transformers import AutoTokenizer
from unilm import UniLMTokenizer
import argparse

def parse_args():
    """Command line argument specification"""
    parser = argparse.ArgumentParser(description="Arguments for Generation Text Detection")

    parser.add_argument(
        "--model_name_or_path",
        type=str,
        default="facebook/opt-1.3b",
        help="Main model, path to pretrained model or model identifier from huggingface.co/models.",
    )
    parser.add_argument(
        "--two_grams_probability_file",
        type=str,
        default="two_grams_probability.json",
        help="The file of 2-grams probability.",
    )
    parser.add_argument(
        "--three_grams_probability_file",
        type=str,
        default="three_grams_probs.json",
        help="The file of 3-grams probability.",
    )
    parser.add_argument(
        "--four_grams_probability_file",
        type=str,
        default="four_grams_probs.json",
        help="The file of 4-grams probability.",
    )
    parser.add_argument(
        "--sample_probability_file",
        type=str,
        default="gpt.npz",
        help="The file of 4-grams probability.",
    )
    parser.add_argument(
        "--test_file",
        type=str,
        default="test_text_3.json",
        help="The result of Proxy Perplexity for test texts.",
    )
    parser.add_argument(
        "--output_path",
        type=str,
        default="gpt_test_ppl.json",
        help="The result of Proxy Perplexity for test texts.",
    )
    args = parser.parse_args()
    return args

# Compress n-grams probabilty
def compress_data(n_grams_probability_file):
    # # Load n-grams probability file
    with open(n_grams_probability_file, 'r', encoding="utf-8") as f:
        grams_probs = json.load(f)

    probability_grams = {}
    two_grams_keys = {}
    two_grams_probs_value = {}
    for i in grams_probs.keys():
        two_grams_keys[tuple(np.uint16(eval(i)))] = np.array(list(grams_probs[i].keys())).astype(np.uint16)
        two_grams_probs_value[tuple(np.uint16(eval(i)))] = np.array(list(grams_probs[i].values())).astype(np.float16)
    probability_grams["keys"] = two_grams_keys
    probability_grams["values"] = two_grams_probs_value


    # probability_2_grams = {}
    # for i in two_grams_probs.keys():
    #     # two_grams_prefix = np.array(eval(i)).astype(np.uint16)   tuple(two_grams_prefix)
    #     two_grams_keys = [np.uint16(k) for k in two_grams_probs[i].keys()]
    #     two_grams_probs_value = [np.float16(v) for v in two_grams_probs[i].values()]
    #     probability_2_grams[i] = dict(zip(two_grams_keys, two_grams_probs_value))


    # Compress 3-grams probability
    # three_grams_prefix = [eval(i) for i in three_grams_probs.keys()]
    # three_grams_prefix = np.array(three_grams_prefix).astype(np.uint16)
    # three_grams_keys = []
    # three_grams_probs_value = []
    # for i in three_grams_probs.keys():
    #     three_grams_keys.append(np.array(list(three_grams_probs[i].keys())).astype(np.uint16))
    #     three_grams_probs_value.append(np.array(list(three_grams_probs[i].values())).astype(np.float16))
    # three_grams_probs_dict = []
    # for keys, values in zip(three_grams_keys, three_grams_probs_value):
    #     three_grams_probs_dict.append({k: v for k, v in zip(keys, values)})
    # probability_3_grams = {tuple(k): v for k, v in zip(three_grams_prefix, three_grams_probs_dict)}

    # probability_3_grams = {}
    # for i in three_grams_probs.keys():
    #     three_grams_prefix = np.array(eval(i)).astype(np.uint16)
    #     three_grams_keys = np.array(list(three_grams_probs[i].keys())).astype(np.uint16)
    #     three_grams_probs_value = np.array(list(three_grams_probs[i].values())).astype(np.float16)
    #     three_grams_probs_dict = {k: v for k, v in zip(three_grams_keys, three_grams_probs_value)}
    #     probability_3_grams[tuple(three_grams_prefix)] = three_grams_probs_dict


    # Compress 4-grams probability
    # four_grams_prefix = [eval(i) for i in four_grams_probs.keys()]
    # four_grams_prefix = np.array(four_grams_prefix).astype(np.uint16)
    # four_grams_keys = []
    # four_grams_probs_value = []
    # for i in four_grams_probs.keys():
    #     four_grams_keys.append(np.array(list(four_grams_probs[i].keys())).astype(np.uint16))
    #     four_grams_probs_value.append(np.array(list(four_grams_probs[i].values())).astype(np.float16))
    # four_grams_probs_dict = []
    # for keys, values in zip(four_grams_keys, four_grams_probs_value):
    #     four_grams_probs_dict.append({k: v for k, v in zip(keys, values)})
    # probability_4_grams = {tuple(k): v for k, v in zip(four_grams_prefix, four_grams_probs_dict)}

    # probability_4_grams = {}
    # for i in four_grams_probs.keys():
    #     four_grams_prefix = np.array(eval(i)).astype(np.uint16)
    #     four_grams_keys = np.array(list(four_grams_probs[i].keys())).astype(np.uint16)
    #     four_grams_probs_value = np.array(list(four_grams_probs[i].values())).astype(np.float16)
    #     four_grams_probs_dict = {k: v for k, v in zip(four_grams_keys, four_grams_probs_value)}
    #     probability_4_grams[tuple(four_grams_prefix)] = four_grams_probs_dict
    #
    # n_grams_probability = [probability_2_grams, probability_3_grams, probability_4_grams]

    # # Save model n-grams probability
    # np.savez(args.sample_probability_file,
    #          two_grams_prefix=n_grams_probability['two_grams_prefix'],
    #          two_grams_keys=n_grams_probability['two_grams_keys'],
    #          two_grams_probs_value=n_grams_probability['two_grams_probability'],
    #          three_grams_prefix=n_grams_probability['three_grams_prefix'],
    #          three_grams_keys=n_grams_probability['three_grams_keys'],
    #          three_grams_probs_value=n_grams_probability['three_grams_probability'],
    #          four_grams_prefix=n_grams_probability['four_grams_prefix'],
    #          four_grams_keys=n_grams_probability['four_grams_keys'],
    #          four_grams_probs_value=n_grams_probability['four_grams_probability'])

    # np.savez(args.sample_probability_file, sample_probs=n_grams_probability)


    return probability_grams

def proxy_perplexity(text_tokenize_ids, n_grams_probability, args):
    # Load n-grams probability
    probability_2_grams_keys = n_grams_probability['sample_probs'][0]["keys"]
    probability_2_grams_values = n_grams_probability['sample_probs'][0]["values"]
    probability_3_grams_keys = n_grams_probability['sample_probs'][1]["keys"]
    probability_3_grams_values = n_grams_probability['sample_probs'][1]["values"]
    probability_4_grams_keys = n_grams_probability['sample_probs'][2]["keys"]
    probability_4_grams_values = n_grams_probability['sample_probs'][2]["values"]

    # Calculate proxy perplexity value for test text set
    test_perplexity = []

    for j in tqdm(range(len(text_tokenize_ids))):
        for label, tokenize_ids in text_tokenize_ids[j].items():
            ppl_result = {}
            ppl = 0
            number_3_grams = 0
            number_4_grams = 0
            number_2_grams = 0
            for i in range(2, len(tokenize_ids) - 1):

                # Calculate the perplexity with 4-grams samples probability
                if tuple([tokenize_ids[i - j] for j in range(2, -1, -1)]) in probability_4_grams_keys.keys():
                    if tokenize_ids[i + 1] in probability_4_grams_keys[tuple([tokenize_ids[i - j] for j in range(2, -1, -1)])]:
                        if probability_4_grams_values[tuple([tokenize_ids[i - j] for j in range(2, -1, -1)])][probability_4_grams_keys[tuple([tokenize_ids[i - j] for j in range(2, -1, -1)])].tolist().index(tokenize_ids[i + 1])] > 0:
                            ppl = ppl + math.log2(probability_4_grams_values[tuple([tokenize_ids[i - j] for j in range(2, -1, -1)])][probability_4_grams_keys[tuple([tokenize_ids[i - j] for j in range(2, -1, -1)])].tolist().index(tokenize_ids[i + 1])])
                    else:
                        top_k = len(probability_4_grams_keys[tuple([tokenize_ids[i - j] for j in range(2, -1, -1)])])
                        sum_probs = sum(probability_4_grams_values[tuple([tokenize_ids[i - j] for j in range(2, -1, -1)])])
                        if (1 - sum_probs) > 0:
                            ppl = ppl + math.log2((1 - sum_probs) / (args.vocab_length - top_k))
                    number_4_grams = number_4_grams + 1

                # Calculate the perplexity with 3-grams samples probability
                elif tuple([tokenize_ids[i - 1], tokenize_ids[i]]) in probability_3_grams_keys.keys():
                    if tokenize_ids[i + 1] in probability_3_grams_keys[tuple([tokenize_ids[i - 1], tokenize_ids[i]])]:
                        if probability_3_grams_values[tuple([tokenize_ids[i - 1], tokenize_ids[i]])][probability_3_grams_keys[tuple([tokenize_ids[i - 1], tokenize_ids[i]])].tolist().index(tokenize_ids[i + 1])] > 0:
                            ppl = ppl + math.log2(probability_3_grams_values[tuple([tokenize_ids[i - 1], tokenize_ids[i]])][probability_3_grams_keys[tuple([tokenize_ids[i - 1], tokenize_ids[i]])].tolist().index(tokenize_ids[i + 1])])
                    else:
                        top_k = len(probability_3_grams_keys[tuple([tokenize_ids[i - 1], tokenize_ids[i]])])
                        sum_probs = sum(probability_3_grams_values[tuple([tokenize_ids[i - 1], tokenize_ids[i]])])
                        if (1 - sum_probs) > 0:
                            ppl = ppl + math.log2((1 - sum_probs) / (args.vocab_length - top_k))
                    number_3_grams = number_3_grams + 1

                # Calculate the perplexity with 2-grams samples probability
                elif tuple([tokenize_ids[i]]) in probability_2_grams_keys.keys():
                    if tokenize_ids[i + 1] in probability_2_grams_keys[tuple([tokenize_ids[i]])]:
                        if probability_2_grams_values[tuple([tokenize_ids[i]])][probability_2_grams_keys[tuple([tokenize_ids[i]])].tolist().index(tokenize_ids[i + 1])] > 0:
                            ppl = ppl + math.log2(probability_2_grams_values[tuple([tokenize_ids[i]])][probability_2_grams_keys[tuple([tokenize_ids[i]])].tolist().index(tokenize_ids[i + 1])])
                    else:
                        top_k = len(probability_2_grams_keys[tuple([tokenize_ids[i]])])
                        sum_probs = sum(probability_2_grams_values[tuple([tokenize_ids[i]])])
                        if (1 - sum_probs) > 0:
                            ppl = ppl + math.log2((1 - sum_probs) / (args.vocab_length - top_k))
                    number_2_grams = number_2_grams + 1

            ppl = round(ppl / (number_2_grams + number_3_grams + number_4_grams + 1), 2)

            ppl_result[label] = -ppl
            test_perplexity.append(ppl_result)

    return test_perplexity

# Proxy perplexity function
def new_proxy_perplexity(text_tokenize_ids, args):
    # Load n-grams probability data
    n_grams_probability = np.load(args.sample_probability_file, allow_pickle=True)

    # 2-grams sample probability
    two_grams_prefix = n_grams_probability['two_grams_prefix']
    two_grams_prefix = [str(i) for i in two_grams_prefix]
    two_grams_keys = n_grams_probability['two_grams_keys']
    # # two_grams_keys = [list(i) for i in two_grams_keys]
    # two_grams_keys = [[str(i) for i in keys] for keys in two_grams_keys]
    two_grams_probability = n_grams_probability['two_grams_probs_value']
    two_grams_probs_dict = []
    for keys, values in zip(two_grams_keys, two_grams_probability):
        two_grams_probs_dict.append({str(k): v for k, v in zip(keys, values)})
    probility_2_grams = {k: v for k, v in zip(two_grams_prefix, two_grams_probs_dict)}

    # 3-grams sample probability
    three_grams_prefix = n_grams_probability['three_grams_prefix']
    three_grams_prefix = [str(i) for i in three_grams_prefix]
    three_grams_keys = n_grams_probability['three_grams_keys']
    # three_grams_keys = [list(i) for i in three_grams_keys]
    # three_grams_keys = [[str(i) for i in keys] for keys in three_grams_keys]
    three_grams_probability = n_grams_probability['three_grams_probs_value']
    three_grams_probs_dict = []
    for keys, values in zip(three_grams_keys, three_grams_probability):
        three_grams_probs_dict.append({str(k): v for k, v in zip(keys, values)})
    probility_3_grams = {k: v for k, v in zip(three_grams_prefix, three_grams_probs_dict)}

    # 4-grams sample probability
    four_grams_prefix = n_grams_probability['four_grams_prefix']
    four_grams_prefix = [str(i) for i in four_grams_prefix]
    four_grams_keys = n_grams_probability['four_grams_keys']
    # four_grams_keys = [list(i) for i in four_grams_keys]
    # four_grams_keys = [[str(i) for i in keys] for keys in four_grams_keys]
    four_grams_probability = n_grams_probability['four_grams_probs_value']
    four_grams_probs_dict = []
    for keys, values in zip(four_grams_keys, four_grams_probability):
        four_grams_probs_dict.append({str(k): v for k, v in zip(keys, values)})
    probility_4_grams = {k: v for k, v in zip(four_grams_prefix, four_grams_probs_dict)}

    # Proxy perplexity value for test text set
    test_ppl = []

    # # Calculate proxy perplexity according to n-grams probability
    # for j in tqdm(range(len(text_tokenize_ids))):
    #     for label, tokenize_ids in text_tokenize_ids[j].items():
    #         ppl_result = {}
    #         ppl = 0
    #         # number_5_grams = 0
    #         number_4_grams = 0
    #         number_3_grams = 0
    #         number_2_grams = 0
    #
    #         for i in range(2, len(tokenize_ids) - 1):
    #
    #             # 4-grams proxy perplexity
    #             if str([tokenize_ids[i - k] for k in range(2, -1, -1)]) in four_grams_prefix:
    #                 prefix_4_grams = str([tokenize_ids[i - k] for k in range(2, -1, -1)])
    #                 key_index = four_grams_prefix.index(prefix_4_grams)
    #                 # if tokenize_ids[i + 1] in four_grams_keys[key_index]:
    #                 # probability_index = four_grams_keys[key_index].index(tokenize_ids[i + 1])
    #                 # # probability_index = np.where(four_grams_keys == tokenize_ids[i + 1])[0][0]
    #                 # probability = four_grams_probability[key_index][probability_index]
    #
    #                 if str(tokenize_ids[i + 1]) in four_grams_probs_dict[key_index].keys():
    #                     probability = four_grams_probs_dict[key_index][str(tokenize_ids[i + 1])]
    #
    #                     # if probability != 0:
    #                     ppl += math.log2(probability)
    #                     number_4_grams += 1
    #                 else:
    #                     # top_k = len(four_grams_keys[key_index])
    #                     # sum_probs = sum(four_grams_probability[key_index])
    #
    #                     top_k = len(four_grams_probs_dict[key_index].keys())
    #                     sum_probs = sum(four_grams_probs_dict[key_index].values())
    #
    #                     if (1 - sum_probs) > 0:
    #                         ppl += math.log2((1 - sum_probs) / (args.vocab_length - top_k))
    #                         number_4_grams += 1
    #
    #             # 3-grams proxy perplexity
    #             elif str([tokenize_ids[i - 1], tokenize_ids[i]]) in three_grams_prefix:
    #                 prefix_3_grams = str([tokenize_ids[i - 1], tokenize_ids[i]])
    #                 key_index = three_grams_prefix.index(prefix_3_grams)
    #
    #                 # if tokenize_ids[i + 1] in three_grams_keys[key_index]:
    #                 #     probability_index = three_grams_keys[key_index].index(tokenize_ids[i + 1])
    #                 #     # probability_index = np.where(three_grams_keys == tokenize_ids[i + 1])[0][0]
    #                 #     probability = three_grams_probability[key_index][probability_index]
    #
    #                 if str(tokenize_ids[i + 1]) in three_grams_probs_dict[key_index].keys():
    #                     probability = three_grams_probs_dict[key_index][str(tokenize_ids[i + 1])]
    #
    #                     # if probability != 0:
    #                     ppl += math.log2(probability)
    #                     number_3_grams += 1
    #                 else:
    #                     # top_k = len(three_grams_keys[key_index])
    #                     # sum_probs = sum(three_grams_probability[key_index])
    #
    #                     top_k = len(three_grams_probs_dict[key_index].keys())
    #                     sum_probs = sum(three_grams_probs_dict[key_index].values())
    #
    #                     if (1 - sum_probs) > 0:
    #                         ppl += math.log2((1 - sum_probs) / (args.vocab_length - top_k))
    #                         number_3_grams += 1
    #
    #             # 2-grams proxy perplexity
    #             elif str([tokenize_ids[i]]) in two_grams_prefix:
    #                 prefix_2_grams = str([tokenize_ids[i]])
    #                 key_index = two_grams_prefix.index(prefix_2_grams)
    #
    #                 # if tokenize_ids[i + 1] in two_grams_keys[key_index]:
    #                 #     probability_index = two_grams_keys[key_index].index(tokenize_ids[i + 1])
    #                 #     # probability_index = np.where(two_grams_keys == tokenize_ids[i + 1])[0][0]
    #                 #     probability = two_grams_probability[key_index][probability_index]
    #
    #                 if str(tokenize_ids[i + 1]) in two_grams_probs_dict[key_index].keys():
    #                     probability = two_grams_probs_dict[key_index][str(tokenize_ids[i + 1])]
    #
    #                     # if probability != 0:
    #                     ppl += math.log2(probability)
    #                     number_2_grams += 1
    #                 else:
    #                     # top_k = len(two_grams_keys[key_index])
    #                     # sum_probs = sum(two_grams_probability[key_index])
    #
    #                     top_k = len(two_grams_probs_dict[key_index].keys())
    #                     sum_probs = sum(two_grams_probs_dict[key_index].values())
    #
    #                     if (1 - sum_probs) > 0:
    #                         ppl += math.log2((1 - sum_probs) / (args.vocab_length - top_k))
    #                         number_2_grams += 1
    #         ppl = round(ppl / (number_2_grams + number_3_grams + number_4_grams + 1), 2)
    #         ppl_result[label] = -ppl
    #         print(-ppl)
    #         test_ppl.append(ppl_result)

    # 代理ppl计算
    for j in tqdm(range(len(text_tokenize_ids))):
        for label, tokenize_ids in text_tokenize_ids[j].items():
            ppl_result = {}
            # ppl = math.log2(probility_2_grams[tokenize_ids[0]][str(tokenize_ids[1])])
            ppl = 0
            number_3_grams = 0
            number_4_grams = 0
            number_2_grams = 0
            for i in range(2, len(tokenize_ids) - 1):

                # 4-grams的代理ppl计算
                if str([tokenize_ids[i - j] for j in range(2, -1, -1)]) in probility_4_grams.keys():
                    if str(tokenize_ids[i + 1]) in probility_4_grams[str([tokenize_ids[i - j] for j in range(2, -1, -1)])].keys():
                        if probility_4_grams[str([tokenize_ids[i - j] for j in range(2, -1, -1)])][str(tokenize_ids[i + 1])] > 0:
                            ppl = ppl + math.log2(probility_4_grams[str([tokenize_ids[i - j] for j in range(2, -1, -1)])][str(tokenize_ids[i + 1])])
                    else:
                        top_k = len(probility_4_grams[str([tokenize_ids[i - j] for j in range(2, -1, -1)])].keys())
                        sum_probs = sum(probility_4_grams[str([tokenize_ids[i - j] for j in range(2, -1, -1)])].values())
                        if (1 - sum_probs) > 0:
                            ppl = ppl + math.log2((1 - sum_probs) / (args.vocab_length - top_k))
                    number_4_grams = number_4_grams + 1

                # 3-grams的代理ppl计算
                elif str([tokenize_ids[i - 1], tokenize_ids[i]]) in probility_3_grams.keys():
                    if str(tokenize_ids[i + 1]) in probility_3_grams[str([tokenize_ids[i - 1], tokenize_ids[i]])].keys():
                        if probility_3_grams[str([tokenize_ids[i - 1], tokenize_ids[i]])][str(tokenize_ids[i + 1])] > 0:
                            ppl = ppl + math.log2(probility_3_grams[str([tokenize_ids[i - 1], tokenize_ids[i]])][str(tokenize_ids[i + 1])])
                    else:
                        top_k = len(probility_3_grams[str([tokenize_ids[i - 1], tokenize_ids[i]])].keys())
                        sum_probs = sum(probility_3_grams[str([tokenize_ids[i - 1], tokenize_ids[i]])].values())
                        if (1 - sum_probs) > 0:
                            ppl = ppl + math.log2((1 - sum_probs) / (args.vocab_length - top_k))
                    number_3_grams = number_3_grams + 1

                elif str([tokenize_ids[i]]) in probility_2_grams.keys():
                    if str(tokenize_ids[i + 1]) in probility_2_grams[str([tokenize_ids[i]])].keys():
                        if probility_2_grams[str([tokenize_ids[i]])][str(tokenize_ids[i + 1])] > 0:
                            ppl = ppl + math.log2(probility_2_grams[str([tokenize_ids[i]])][str(tokenize_ids[i + 1])])
                    else:
                        top_k = len(probility_2_grams[str([tokenize_ids[i]])].keys())
                        sum_probs = sum(probility_2_grams[str([tokenize_ids[i]])].values())
                        ppl = ppl + math.log2((1 - sum_probs) / (args.vocab_length - top_k))
                    number_2_grams = number_2_grams + 1

            # 没有一个全局2grams情况下的算法
            ppl = round(ppl / (number_2_grams + number_3_grams + number_4_grams + 1), 2)

            ppl_result[label] = -ppl
            test_ppl.append(ppl_result)

    return test_ppl

def main(args):
    """Run a command line version of the generation and detection operations
        and optionally launch and serve the gradio demo"""
    # print(args)
    args.is_opt_model = any([(model_type in args.model_name_or_path) for model_type in ["opt"]])
    args.is_gpt_model = any([(model_type in args.model_name_or_path) for model_type in ["gpt"]])
    args.is_unilm_model = any([(model_type in args.model_name_or_path) for model_type in ["unilm"]])
    if args.is_opt_model:
        args.model_name = "opt"
        args.vocab_length = 50265
    if args.is_gpt_model:
        args.model_name = "gpt"
        args.vocab_length = 50257
    if args.is_unilm_model:
        args.model_name = "unilm"
        args.vocab_length = 28996

    if args.is_unilm_model:
        tokenizer = UniLMTokenizer.from_pretrained(args.model_name_or_path)
    else:
        tokenizer = AutoTokenizer.from_pretrained(args.model_name_or_path)

    test_text = []
    with open(args.test_file, 'r', encoding="utf-8") as f:
        for line in f:
            try:
                text = json.loads(line)
                test_text.append(text)
            except Exception as e:
                print(e, line)
    # test_text = test_text[0:100]

    test_text_tokenize_ids = []
    for text in test_text:
        tokenize_ids = {}
        if args.is_opt_model:
            # 对text进行tokenize
            text_tokenize = tokenizer.tokenize(text["text"])
        elif args.is_gpt_model:
            # 对text进行tokenize
            text_tokenize = tokenizer.tokenize(text["text"], truncation=True)
        elif args.is_unilm_model:
            text_tokenize = tokenizer.tokenize(text["text"])
        tokenize_ids[text["label"]] = tokenizer.convert_tokens_to_ids(text_tokenize)
        # tokenize转token_ids
        test_text_tokenize_ids.append(tokenize_ids)

    # Load n-grams probability
    n_grams_probability = np.load(args.sample_probability_file, allow_pickle=True)
    # test_ppl = new_proxy_perplexity(test_text_tokenize_ids, n_grams_probability, args)
    test_ppl = proxy_perplexity(test_text_tokenize_ids, n_grams_probability, args)
    with open(args.output_path, 'w', encoding='utf-8') as f:
        for i in range(len(test_ppl)):
            f.write(json.dumps(test_ppl[i], ensure_ascii=False))
            f.write('\n')

if __name__ == "__main__":
    args = parse_args()
    # print(args)
    probability_2_grams = compress_data(args.two_grams_probability_file)
    probability_3_grams = compress_data(args.three_grams_probability_file)
    probability_4_grams = compress_data(args.four_grams_probability_file)

    n_grams_probability = [probability_2_grams, probability_3_grams, probability_4_grams]
    # n_grams_probability = [probability_2_grams]

    np.savez(args.sample_probability_file, sample_probs=n_grams_probability)

    main(args)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="llmdet",
    version="1.0.3",
    author="Kangxi Wu, Liang Pang",
    author_email="wukx0901@gmail.com",
    description="LLMDet：A Large Language Models Detection Tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cansee5/LLMDet",
    packages=setuptools.find_packages(),
    install_requires=[
            'transformers',
            'nltk',
            'lightgbm',
            'numpy',
            'sklearn',
            'tqdm',
            'argparse',
        ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



import numpy as np
import datasets
from tqdm import tqdm
import math
import json
from transformers import AutoTokenizer
from unilm import UniLMTokenizer
from lightgbm import Booster


def load_probability():
    dm = datasets.DownloadManager()
    files = dm.download_and_extract('https://huggingface.co/datasets/wukx/n-grams_sample_probability/resolve/main/new_n_grams.zip')
    n_grams = np.load(f'{files}/new_n_grams.npz' , allow_pickle=True)
    global gpt_n_grams_probability, opt_n_grams_probability, unilm_n_grams_probability
    gpt_n_grams_probability = n_grams["gpt"]
    opt_n_grams_probability = n_grams["opt"]
    unilm_n_grams_probability = n_grams["unilm"]

# Calculate the perplexity of text.
def perplexity(text_set_token_ids, n_grams_probability, vocab_size):

    # Calculate proxy perplexity value for test text set
    test_perplexity = []

    for k in tqdm(range(len(text_set_token_ids))):
        text_token_ids = text_set_token_ids[k]
        ppl = 0
        number_3_grams = 0
        number_4_grams = 0
        number_2_grams = 0
        for i in range(2, len(text_token_ids) - 1):

            # Calculate the perplexity with 4-grams samples probability
            if tuple([text_token_ids[i - j] for j in range(2, -1, -1)]) in n_grams_probability[4].keys():
                if text_token_ids[i + 1] in n_grams_probability[4][tuple([text_token_ids[i - j] for j in range(2, -1, -1)])]:
                    if n_grams_probability[5][tuple([text_token_ids[i - j] for j in range(2, -1, -1)])][n_grams_probability[4][tuple([text_token_ids[i - j] for j in range(2, -1, -1)])].tolist().index(text_token_ids[i + 1])] > 0:
                        ppl = ppl + math.log2(n_grams_probability[5][tuple([text_token_ids[i - j] for j in range(2, -1, -1)])][n_grams_probability[4][tuple([text_token_ids[i - j] for j in range(2, -1, -1)])].tolist().index(text_token_ids[i + 1])])
                else:
                    top_k = len(n_grams_probability[4][tuple([text_token_ids[i - j] for j in range(2, -1, -1)])])
                    sum_probs = sum(n_grams_probability[5][tuple([text_token_ids[i - j] for j in range(2, -1, -1)])])
                    if (1 - sum_probs) > 0:
                        ppl = ppl + math.log2((1 - sum_probs) / (vocab_size - top_k))
                number_4_grams = number_4_grams + 1

            # Calculate the perplexity with 3-grams samples probability
            elif tuple([text_token_ids[i - 1], text_token_ids[i]]) in n_grams_probability[2].keys():
                if text_token_ids[i + 1] in n_grams_probability[2][tuple([text_token_ids[i - 1], text_token_ids[i]])]:
                    if n_grams_probability[3][tuple([text_token_ids[i - 1], text_token_ids[i]])][n_grams_probability[2][tuple([text_token_ids[i - 1], text_token_ids[i]])].tolist().index(text_token_ids[i + 1])] > 0:
                        ppl = ppl + math.log2(
                            n_grams_probability[3][tuple([text_token_ids[i - 1], text_token_ids[i]])][n_grams_probability[2][tuple([text_token_ids[i - 1], text_token_ids[i]])].tolist().index(text_token_ids[i + 1])])
                else:
                    top_k = len(n_grams_probability[2][tuple([text_token_ids[i - 1], text_token_ids[i]])])
                    sum_probs = sum(n_grams_probability[3][tuple([text_token_ids[i - 1], text_token_ids[i]])])
                    if (1 - sum_probs) > 0:
                        ppl = ppl + math.log2((1 - sum_probs) / (vocab_size - top_k))
                number_3_grams = number_3_grams + 1

            # Calculate the perplexity with 2-grams samples probability
            elif tuple([text_token_ids[i]]) in n_grams_probability[0].keys():
                if text_token_ids[i + 1] in n_grams_probability[0][tuple([text_token_ids[i]])]:
                    if n_grams_probability[1][tuple([text_token_ids[i]])][n_grams_probability[0][tuple([text_token_ids[i]])].tolist().index(text_token_ids[i + 1])] > 0:
                        ppl = ppl + math.log2(n_grams_probability[1][tuple([text_token_ids[i]])][n_grams_probability[0][tuple([text_token_ids[i]])].tolist().index(text_token_ids[i + 1])])
                else:
                    top_k = len(n_grams_probability[0][tuple([text_token_ids[i]])])
                    sum_probs = sum(n_grams_probability[1][tuple([text_token_ids[i]])])
                    if (1 - sum_probs) > 0:
                        ppl = ppl + math.log2((1 - sum_probs) / (vocab_size - top_k))
                number_2_grams = number_2_grams + 1

        perplexity = round(ppl / (number_2_grams + number_3_grams + number_4_grams + 1), 2)
        test_perplexity.append(-perplexity)

    return test_perplexity

# Detect function
def detect(text):
    # Determine whether the input is a single text or a collection of text.
    if isinstance(text, str):
        test_text = [text]
    elif isinstance(text, list):
        if isinstance(text[0], str):
            test_text = text
        else:
            raise ValueError(
                "The type of `text` which you input is not a string or list. "
                "Please enter the correct data type for `text`."
            )
    else:
        raise ValueError(
            "The type of `text` which you input is not a string or list. "
            "Please enter the correct data type for `text`."
        )
    dm = datasets.DownloadManager()
    # files = dm.download_and_extract('https://huggingface.co/datasets/wukx/n-grams_sample_probability/resolve/main/n_grams.zip')

    model_information = [{"model_name": "gpt2", "vocab_size": 50265, "model_probability": "gpt_n_grams_probability"},
                         {"model_name": "facebook/opt-1.3b", "vocab_size": 50257, "model_probability": "opt_n_grams_probability"},
                         {"model_name": "microsoft/unilm-base-cased", "vocab_size": 28996, "model_probability": "unilm_n_grams_probability"}]

    # Calculate the proxy perplexity
    perplexity_result = []
    for model in model_information:
        if any([(model_type in model["model_name"]) for model_type in ["unilm"]]):
            tokenizer = UniLMTokenizer.from_pretrained(model["model_name"])
        else:
            tokenizer = AutoTokenizer.from_pretrained(model["model_name"])

        text_token_ids = []
        for text in test_text:
            if any([(model_type in model["model_name"]) for model_type in ["gpt"]]):
                text_tokenize = tokenizer.tokenize(text, truncation=True)
            else:
                text_tokenize = tokenizer.tokenize(text)
            text_token_ids.append(tokenizer.convert_tokens_to_ids(text_tokenize))


        if model["model_probability"] in globals():
            perplexity_result.append(perplexity(text_token_ids, globals()[model["model_probability"]], model["vocab_size"]))
        else:
            raise ValueError(
                "The {} does not exist, please load n-grams probability!".format(model["model_probability"])
            )

    # The input features of classiffier
    features = np.stack([perplexity_result[i] for i in range(len(perplexity_result))], axis=1)
    print(features)

    # Load classiffier model
    model_files = dm.download_and_extract(
        'https://huggingface.co/datasets/wukx/n-grams_sample_probability/resolve/main/LightGBM_model.zip')
    model = Booster(model_file=f'{model_files}/LightGBM_model.txt')
    y_pred = model.predict(features)
    label = ["Human_write", "GPT-2", "OPT", "UniLM"]
    test_result = [{label[i]: y_pred[j][i] for i in range(len(label))} for j in range(len(y_pred))]

    return test_result
